"""ClearingFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract


from .types import Trade


class ClearingFacet:
    """ClearingFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed ClearingFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    @property
    def FeeCollected(self) -> contract.ContractEvent:
        """Binding for `event FeeCollected` on the ClearingFacet contract."""
        return self._contract.events.FeeCollected

    @property
    def FeeDispersed(self) -> contract.ContractEvent:
        """Binding for `event FeeDispersed` on the ClearingFacet contract."""
        return self._contract.events.FeeDispersed

    @property
    def PositionUpdated(self) -> contract.ContractEvent:
        """Binding for `event PositionUpdated` on the ClearingFacet contract."""
        return self._contract.events.PositionUpdated

    @property
    def TradeExecuted(self) -> contract.ContractEvent:
        """Binding for `event TradeExecuted` on the ClearingFacet contract."""
        return self._contract.events.TradeExecuted

    def estimate_fees(
        self,
        product_id: hexbytes.HexBytes,
        price: int,
        quantity: int,
        trading_fee_rate: int,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> typing.Tuple[int, int]:
        """Binding for `estimateFees` on the ClearingFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        price : int
        quantity : int
        trading_fee_rate : int
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        int
        """
        return_value = self._contract.functions.estimateFees(
            product_id,
            price,
            quantity,
            trading_fee_rate,
        ).call(block_identifier=block_identifier)
        return (
            int(return_value[0]),
            int(return_value[1]),
        )

    def execute(
        self,
        trade: Trade,
        fallback_on_failure: bool,
    ) -> contract.ContractFunction:
        """Binding for `execute` on the ClearingFacet contract.

        Parameters
        ----------
        trade : Trade
        fallback_on_failure : bool

        Returns
        -------
        web3.contract.contract.ContractFunction
            A contract function instance to be sent in a transaction.
        """
        return self._contract.functions.execute(
            (
                trade.product_id,
                trade.protocol_id,
                trade.trade_id,
                trade.price,
                trade.timestamp,
                trade.quantity,
                trade.accounts,
                trade.quantities,
                trade.fee_rates,
                [
                    (
                        trade_item.margin_account_id,
                        trade_item.intent_account_id,
                        trade_item.hash,
                        (
                            trade_item.data.nonce,
                            trade_item.data.trading_protocol_id,
                            trade_item.data.product_id,
                            trade_item.data.limit_price,
                            trade_item.data.quantity,
                            trade_item.data.max_trading_fee_rate,
                            trade_item.data.good_until,
                            int(trade_item.data.side),
                            trade_item.data.referral,
                        ),
                        trade_item.signature,
                    )
                    for trade_item in trade.intents
                ],
            ),
            fallback_on_failure,
        )


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "estimateFees",
            "inputs": [
                {"name": "productID", "type": "bytes32", "internalType": "bytes32"},
                {"name": "price", "type": "int256", "internalType": "int256"},
                {"name": "quantity", "type": "uint256", "internalType": "uint256"},
                {"name": "tradingFeeRate", "type": "int256", "internalType": "int256"},
            ],
            "outputs": [
                {"name": "", "type": "uint256", "internalType": "uint256"},
                {"name": "", "type": "int256", "internalType": "int256"},
            ],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "execute",
            "inputs": [
                {
                    "name": "trade",
                    "type": "tuple",
                    "internalType": "struct Trade",
                    "components": [
                        {
                            "name": "productId",
                            "type": "bytes32",
                            "internalType": "bytes32",
                        },
                        {
                            "name": "protocolId",
                            "type": "address",
                            "internalType": "address",
                        },
                        {
                            "name": "tradeId",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {"name": "price", "type": "int256", "internalType": "int256"},
                        {
                            "name": "timestamp",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "quantity",
                            "type": "uint256",
                            "internalType": "uint256",
                        },
                        {
                            "name": "accounts",
                            "type": "address[]",
                            "internalType": "address[]",
                        },
                        {
                            "name": "quantities",
                            "type": "uint256[]",
                            "internalType": "uint256[]",
                        },
                        {
                            "name": "feeRates",
                            "type": "int32[]",
                            "internalType": "int32[]",
                        },
                        {
                            "name": "intents",
                            "type": "tuple[]",
                            "internalType": "struct Intent[]",
                            "components": [
                                {
                                    "name": "marginAccountId",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "intentAccountId",
                                    "type": "address",
                                    "internalType": "address",
                                },
                                {
                                    "name": "hash",
                                    "type": "bytes32",
                                    "internalType": "bytes32",
                                },
                                {
                                    "name": "data",
                                    "type": "tuple",
                                    "internalType": "struct IntentData",
                                    "components": [
                                        {
                                            "name": "nonce",
                                            "type": "uint256",
                                            "internalType": "uint256",
                                        },
                                        {
                                            "name": "tradingProtocolId",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                        {
                                            "name": "productId",
                                            "type": "bytes32",
                                            "internalType": "bytes32",
                                        },
                                        {
                                            "name": "limitPrice",
                                            "type": "int256",
                                            "internalType": "int256",
                                        },
                                        {
                                            "name": "quantity",
                                            "type": "uint256",
                                            "internalType": "uint256",
                                        },
                                        {
                                            "name": "maxTradingFeeRate",
                                            "type": "uint32",
                                            "internalType": "uint32",
                                        },
                                        {
                                            "name": "goodUntil",
                                            "type": "uint256",
                                            "internalType": "uint256",
                                        },
                                        {
                                            "name": "side",
                                            "type": "uint8",
                                            "internalType": "enum Side",
                                        },
                                        {
                                            "name": "referral",
                                            "type": "address",
                                            "internalType": "address",
                                        },
                                    ],
                                },
                                {
                                    "name": "signature",
                                    "type": "bytes",
                                    "internalType": "bytes",
                                },
                            ],
                        },
                    ],
                },
                {"name": "fallbackOnFailure", "type": "bool", "internalType": "bool"},
            ],
            "outputs": [],
            "stateMutability": "nonpayable",
        },
        {
            "type": "event",
            "name": "FeeCollected",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "capitalAmount",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "FeeDispersed",
            "inputs": [
                {
                    "name": "recipient",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "collateralAsset",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "capitalAmount",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "PositionUpdated",
            "inputs": [
                {
                    "name": "marginAccountId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "productId",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "costBasis",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "price",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "quantity",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "event",
            "name": "TradeExecuted",
            "inputs": [
                {
                    "name": "productId",
                    "type": "bytes32",
                    "indexed": True,
                    "internalType": "bytes32",
                },
                {
                    "name": "protocolId",
                    "type": "address",
                    "indexed": True,
                    "internalType": "address",
                },
                {
                    "name": "id",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
                {
                    "name": "price",
                    "type": "int256",
                    "indexed": False,
                    "internalType": "int256",
                },
                {
                    "name": "quantity",
                    "type": "uint256",
                    "indexed": False,
                    "internalType": "uint256",
                },
            ],
            "anonymous": False,
        },
        {
            "type": "error",
            "name": "DuplicateMarginAccount",
            "inputs": [
                {"name": "marginAccount", "type": "address", "internalType": "address"}
            ],
        },
        {"type": "error", "name": "ECDSAInvalidSignature", "inputs": []},
        {
            "type": "error",
            "name": "ECDSAInvalidSignatureLength",
            "inputs": [
                {"name": "length", "type": "uint256", "internalType": "uint256"}
            ],
        },
        {
            "type": "error",
            "name": "ECDSAInvalidSignatureS",
            "inputs": [{"name": "s", "type": "bytes32", "internalType": "bytes32"}],
        },
        {
            "type": "error",
            "name": "IntentFullySpent",
            "inputs": [
                {"name": "intentAccount", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidFeeSum",
            "inputs": [{"name": "feeSum", "type": "int256", "internalType": "int256"}],
        },
        {
            "type": "error",
            "name": "InvalidFieldAccess",
            "inputs": [
                {
                    "name": "productType",
                    "type": "uint8",
                    "internalType": "enum ProductType",
                },
                {"name": "field", "type": "string", "internalType": "string"},
            ],
        },
        {
            "type": "error",
            "name": "InvalidIntent",
            "inputs": [
                {"name": "parameter", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidParameter",
            "inputs": [
                {"name": "paramName", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidProductState",
            "inputs": [
                {"name": "state", "type": "uint8", "internalType": "enum ProductState"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidSignature",
            "inputs": [
                {"name": "signer", "type": "address", "internalType": "address"},
                {
                    "name": "expectedSigner",
                    "type": "address",
                    "internalType": "address",
                },
            ],
        },
        {
            "type": "error",
            "name": "InvalidTradeIntents",
            "inputs": [
                {"name": "length", "type": "uint256", "internalType": "uint256"}
            ],
        },
        {
            "type": "error",
            "name": "InvalidTradePrice",
            "inputs": [{"name": "price", "type": "int256", "internalType": "int256"}],
        },
        {
            "type": "error",
            "name": "MAECheckFailed",
            "inputs": [
                {"name": "marginAccount", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "MismatchedTrade",
            "inputs": [
                {"name": "buySide", "type": "uint256", "internalType": "uint256"},
                {"name": "sellSide", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "NotFound",
            "inputs": [
                {"name": "parameter", "type": "string", "internalType": "string"}
            ],
        },
        {
            "type": "error",
            "name": "NotImplemented",
            "inputs": [{"name": "feature", "type": "string", "internalType": "string"}],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv18_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
                {"name": "denominator", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Div_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Div_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp2_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Mul_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Mul_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
        {
            "type": "error",
            "name": "SafeCastOverflowedUintToInt",
            "inputs": [{"name": "value", "type": "uint256", "internalType": "uint256"}],
        },
        {
            "type": "error",
            "name": "Unauthorized",
            "inputs": [
                {"name": "account", "type": "address", "internalType": "address"}
            ],
        },
        {
            "type": "error",
            "name": "UnauthorizedTradeSubmitter",
            "inputs": [
                {"name": "needed", "type": "address", "internalType": "address"},
                {"name": "got", "type": "address", "internalType": "address"},
            ],
        },
    ],
)



import io
import json
import os
from typing import Any, cast

import dag_cbor
import ipld_car  # type: ignore (untyped library)
import multiformats
import requests
from requests import Response

from . import constants, types
from .dtos import ComponentLink, ExtendedMetadata, ExtendedMetadataDAG
from .exceptions import IPFSError, ValidationError
from .schemas import OracleConfig, OracleFallback, OutcomePoint, OutcomeSpace
from .types import Model, PinnedModel


class IPFSClient:
    _api_url: str
    _api_key: str | None

    def __init__(self, api_url: str, api_key: str | None = None):
        self._api_url = api_url
        self._api_key = api_key

    def upload_extended_metadata(
        self, extended_metadata: ExtendedMetadata
    ) -> types.CID:
        """Uploads extended metadata as a single CAR file."""

        # Use mode="json" to convert Decimal & datetime types to strings
        outcome_space_cid, outcome_space_data = self.encode(
            extended_metadata.outcome_space.model_dump(mode="json")
        )
        outcome_point_cid, outcome_point_data = self.encode(
            extended_metadata.outcome_point.model_dump(mode="json")
        )
        oracle_config_cid, oracle_config_data = self.encode(
            extended_metadata.oracle_config.model_dump(mode="json")
        )
        oracle_fallback_cid, oracle_fallback_data = self.encode(
            extended_metadata.oracle_fallback.model_dump(mode="json")
        )
        outcome_space_schema_cid, outcome_space_schema_data = self.encode(
            self._load_schema_json(extended_metadata.outcome_space.SCHEMA_CID)
        )
        outcome_point_schema_cid, outcome_point_schema_data = self.encode(
            self._load_schema_json(extended_metadata.outcome_point.SCHEMA_CID)
        )
        oracle_config_schema_cid, oracle_config_schema_data = self.encode(
            self._load_schema_json(extended_metadata.oracle_config.SCHEMA_CID)
        )
        oracle_fallback_schema_cid, oracle_fallback_schema_data = self.encode(
            self._load_schema_json(extended_metadata.oracle_fallback.SCHEMA_CID)
        )

        self.ensure_cids_match(
            outcome_space_schema_cid, extended_metadata.outcome_space.SCHEMA_CID
        )
        self.ensure_cids_match(
            outcome_point_schema_cid, extended_metadata.outcome_point.SCHEMA_CID
        )
        self.ensure_cids_match(
            oracle_config_schema_cid, extended_metadata.oracle_config.SCHEMA_CID
        )
        self.ensure_cids_match(
            oracle_fallback_schema_cid, extended_metadata.oracle_fallback.SCHEMA_CID
        )

        extended_metadata_dag = ExtendedMetadataDAG(
            outcome_space=ComponentLink(
                data=str(outcome_space_cid),
                schema_=str(outcome_space_schema_cid),
            ),
            outcome_point=ComponentLink(
                data=str(outcome_point_cid),
                schema_=str(outcome_point_schema_cid),
            ),
            oracle_config=ComponentLink(
                data=str(oracle_config_cid),
                schema_=str(oracle_config_schema_cid),
            ),
            oracle_fallback=ComponentLink(
                data=str(oracle_fallback_cid),
                schema_=str(oracle_fallback_schema_cid),
            ),
        )
        # Use mode="python" to preserve multiformats.CID types so that the DAG-CBOR
        # encoder will convert them into IPLD Link format
        extended_metadata_dag_cid, extended_metadata_dag_data = self.encode(
            extended_metadata_dag.model_dump(mode="python")
        )

        blocks: list[ipld_car.Block] = [
            (extended_metadata_dag_cid, extended_metadata_dag_data),
            (outcome_space_cid, outcome_space_data),
            (outcome_point_cid, outcome_point_data),
            (oracle_config_cid, oracle_config_data),
            (oracle_fallback_cid, oracle_fallback_data),
            (outcome_space_schema_cid, outcome_space_schema_data),
            (outcome_point_schema_cid, outcome_point_schema_data),
            (oracle_config_schema_cid, oracle_config_schema_data),
            (oracle_fallback_schema_cid, oracle_fallback_schema_data),
        ]
        root_cid = self.upload_car(blocks)
        self.ensure_cids_match(root_cid, extended_metadata_dag_cid)
        return root_cid

    def download_extended_metadata(self, cid: types.CID) -> ExtendedMetadata:
        """Downloads and assembles extended metadata objects."""

        extended_metadata_dag = self._download_and_validate_block(
            cid, ExtendedMetadataDAG
        )

        outcome_space_model = cast(
            type[OutcomeSpace],
            self._find_model_by_schema_cid(extended_metadata_dag.outcome_space.schema_),
        )
        outcome_point_model = cast(
            type[OutcomePoint],
            self._find_model_by_schema_cid(extended_metadata_dag.outcome_point.schema_),
        )
        oracle_config_model = cast(
            type[OracleConfig],
            self._find_model_by_schema_cid(extended_metadata_dag.oracle_config.schema_),
        )
        oracle_fallback_model = cast(
            type[OracleFallback],
            self._find_model_by_schema_cid(
                extended_metadata_dag.oracle_fallback.schema_
            ),
        )

        return ExtendedMetadata(
            outcome_space=self._download_and_validate_block(
                extended_metadata_dag.outcome_space.data, outcome_space_model
            ),
            outcome_point=self._download_and_validate_block(
                extended_metadata_dag.outcome_point.data, outcome_point_model
            ),
            oracle_config=self._download_and_validate_block(
                extended_metadata_dag.oracle_config.data, oracle_config_model
            ),
            oracle_fallback=self._download_and_validate_block(
                extended_metadata_dag.oracle_fallback.data, oracle_fallback_model
            ),
        )

    def upload_car(self, blocks: list[ipld_car.Block]) -> types.CID:
        root_cid = blocks[0][0]
        data = ipld_car.encode([root_cid], blocks).tobytes()

        response = self._send_client_request(
            "/api/v0/dag/import",
            params={"pin-roots": "true"},
            files={"file": ("data.car", io.BytesIO(data), "application/vnd.ipld.car")},
        )

        try:
            response_data = response.json()
        except json.JSONDecodeError as json_error:
            raise IPFSError(
                f"Error decoding IPFS response: {response.text}"
            ) from json_error

        try:
            return response_data["Root"]["Cid"]["/"]
        except (TypeError, KeyError):
            raise IPFSError(f"Unexpected IPFS response format: {response_data}")

    def _download_and_validate_block[T: Model](
        self, cid: types.CID, model: type[T]
    ) -> T:
        codec_name = multiformats.CID.decode(cid).codec.name
        response = self._send_client_request("/api/v0/block/get", params={"arg": cid})

        if codec_name == "dag-cbor":
            return model.model_validate(dag_cbor.decode(response.content))
        elif codec_name == "dag-json":
            return model.model_validate_json(response.content)
        else:
            raise IPFSError(f"Unsupported codec: {codec_name}")

    def _send_client_request(self, endpoint: str, **kwargs: Any) -> Response:
        headers: dict[str, str] = {}
        if self._api_key:
            headers["Authorization"] = f"Bearer {self._api_key}"

        url = f"{self._api_url}{endpoint}"
        try:
            response = requests.post(
                url,
                headers=headers,
                timeout=constants.IPFS_REQUEST_TIMEOUT,
                **kwargs,
            )
        except requests.exceptions.Timeout as timeout_error:
            raise IPFSError(
                f"Request to IPFS client timed out after "
                f"{constants.IPFS_REQUEST_TIMEOUT}s: {url}"
            ) from timeout_error
        except requests.exceptions.RequestException as request_exception:
            raise IPFSError(
                f"Failed to send request to IPFS client: {url}"
            ) from request_exception

        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as http_error:
            raise IPFSError(
                f"IPFS client returned HTTP {response.status_code}: {url}"
            ) from http_error

        return response

    @staticmethod
    def encode(value: Any) -> tuple[multiformats.CID, bytes]:
        cbor_data = dag_cbor.encode(value)
        digest = multiformats.multihash.digest(cbor_data, "sha2-256")
        cid = multiformats.CID(constants.IPFS_CID_ENCODING, 1, "dag-cbor", digest)
        return (cid, cbor_data)

    @staticmethod
    def ensure_cids_match(
        actual_cid: types.CID | multiformats.CID,
        computed_cid: types.CID | multiformats.CID,
    ) -> None:
        if str(computed_cid) != str(actual_cid):
            raise IPFSError(
                f"Mismatch between computed CID '{computed_cid}' "
                f"and actual CID '{actual_cid}'"
            )

    @staticmethod
    def _find_model_by_schema_cid(cid: types.CID) -> type[PinnedModel]:
        if cid not in types.CID_MODEL_MAP:
            raise ValidationError(f"Unsupported schema CID: {cid}")
        return types.CID_MODEL_MAP[cid]

    @staticmethod
    def _load_schema_json(cid: types.CID) -> dict[Any, Any]:
        with open(os.path.join(constants.JSON_SCHEMAS_DIRECTORY, f"{cid}.json")) as f:
            return json.load(f)



from datetime import datetime
from decimal import Decimal
from typing import cast

import eth_account
from eth_typing.evm import ChecksumAddress
from hexbytes import HexBytes

from afp import hashing
from afp.enums import OrderSide
from afp.schemas import IntentData


def test_generate_intent_hash():
    intent_data = IntentData(
        trading_protocol_id="0x783b16190c71278E78f69Af32815CcA818A9822f",
        product_id="0xa74b5c2da3083e4189f56b2582c4455d21bbbd199d7699fc6ebddd82e8fd02b5",
        limit_price=Decimal("1.0"),
        quantity=1,
        max_trading_fee_rate=Decimal("0.1"),
        side=OrderSide.ASK,
        good_until_time=datetime.fromisoformat("2030-01-01T12:00:00Z"),
        nonce=42,
        referral="0x0000000000000000000000000000000000000000",
    )
    margin_account_id = cast(
        ChecksumAddress, "0x51541B823f9C28e8E43E18c0F0d50B405FCB8aD4"
    )
    intent_account_id = cast(
        ChecksumAddress, "0x44163AaD36b0f6986fd09D0d6AD990F0341F16E7"
    )

    expected_hash = HexBytes(
        "0xed1d85647d3cbd0f2ad6bfdbbfcec5c4d1226ecf1bb6579c10cced7d245e1de2"
    )
    actual_hash = hashing.generate_intent_hash(
        intent_data=intent_data,
        margin_account_id=margin_account_id,
        intent_account_id=intent_account_id,
        tick_size=18,
    )
    assert actual_hash == expected_hash


def test_generate_order_cancellation_hash():
    intent_hash = "0x9540b615aa488fb276c71231926d81b2d5f1d43711d306c630df1cf86c81ed88"
    nonce = 42

    expected_hash = HexBytes(
        "0x95672a703e451a4617ecd229400b675e59d21c0d866325cdc87a86d84a898fe8"
    )
    actual_hash = hashing.generate_order_cancellation_hash(nonce, intent_hash)
    assert actual_hash == expected_hash


def test_generate_product_id():
    builder = eth_account.Account.from_key(
        "0x0cc549714a138639cbd51f79ff38f279e490afd2b2b98cc910af6e3d699d6049"
    )
    symbol = "PROD-X-1"

    expected_hash = HexBytes(
        "0x89ed228125682fca3832f5fb52236ebabc3d29e64fa0867556a0621a310c49d7"
    )
    actual_hash = hashing.generate_product_id(builder.address, symbol)
    assert actual_hash == expected_hash



import decimal
from datetime import date, datetime, UTC
from decimal import Decimal

import pytest

from afp import validators
from afp.schemas import Model, Timestamp


class DummyModel(Model):
    timestamp: Timestamp


def test_timestamp_conversion():
    dt_time = datetime.fromisoformat("2030-01-01T12:00:00Z")
    ts_time = 1893499200

    # timestamp -> datetime
    instance = DummyModel(timestamp=ts_time)  # type: ignore
    assert instance.timestamp is not None
    assert instance.timestamp.astimezone(UTC) == dt_time

    # datetime -> timestamp
    assert '"timestamp":%d' % ts_time in instance.model_dump_json()


def test_validate_hexstr32__pass():
    validators.validate_hexstr32(
        "0xe50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b"
    )


@pytest.mark.parametrize(
    "value",
    [
        "e50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b",
        "0xg50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb61b",
        "0xe50c0a9639bdec3c05484a4e912650e63039fd5032f4050b1d1cdd0dd0efb6",
    ],
    ids=str,
)
def test_validate_hexstr32__error(value):
    with pytest.raises(ValueError):
        validators.validate_hexstr32(value)


def test_validate_limit_price__pass():
    assert (
        str(
            validators.validate_limit_price(
                Decimal("1.25"), Decimal("1"), Decimal("2"), 2
            )
        )
        == "1.25"
    )
    assert (
        str(
            validators.validate_limit_price(
                Decimal("1.25"), Decimal("1"), Decimal("2"), 4
            )
        )
        == "1.2500"
    )


def test_validate_limit_price__rounding():
    assert (
        str(
            validators.validate_limit_price(
                Decimal("1.25"), Decimal("1"), Decimal("2"), 1, decimal.ROUND_DOWN
            )
        )
        == "1.2"
    )


def test_validate_limit_price__not_enough_digits():
    with pytest.raises(ValueError):
        validators.validate_limit_price(Decimal("1.25"), Decimal("1"), Decimal("2"), 1)


def test_validate_limit_price__invalid_rounding_mode():
    with pytest.raises(TypeError):
        validators.validate_limit_price(
            Decimal("1.25"), Decimal("1"), Decimal("2"), 1, "foobar"
        )


def test_validate_limit_price__min_price_breach():
    with pytest.raises(ValueError):
        validators.validate_limit_price(
            Decimal("1.25"), Decimal("1.26"), Decimal("2"), 2
        )


def test_validate_limit_price__max_price_breach():
    with pytest.raises(ValueError):
        validators.validate_limit_price(
            Decimal("1.25"), Decimal("1"), Decimal("1.24"), 2
        )


def test_validate_price_limits__pass():
    validators.validate_price_limits(Decimal("0.11"), Decimal("0.12"))


def test_validate_price_limits__error():
    with pytest.raises(ValueError):
        validators.validate_price_limits(Decimal("0.11"), Decimal("0.10"))


def test_validate_outcome_space_template_variables__pass():
    dct = {
        "a": {
            "b": {
                "c": "hello",
                "d": "world",
            },
        },
    }

    validators.validate_outcome_space_template_variables(
        [
            "Reference to {a.b.c} should pass",
            "And to {a.b.d} as well",
            "So as having no template variable",
        ],
        dct,
    )


@pytest.mark.parametrize(
    "value",
    [
        "{a.b.c} and {a.b.e}",
        "{c}",
        "{a.b.c.d}",
        "{a.b}",
        "{}",
    ],
    ids=str,
)
def test_validate_outcome_space_tempate_variables__error(value):
    dct = {
        "a": {
            "b": {
                "c": "hello",
                "d": "world",
            },
        },
    }

    with pytest.raises(ValueError):
        validators.validate_outcome_space_template_variables([value], dct)


@pytest.mark.parametrize(
    "symbol,frequency,release_date",
    [
        ("BTCUSD15M24", "daily", "2024-06-15"),
        ("CLAIMS03W25", "weekly", "2025-01-13"),
        ("PAYROLL02W25", "fortnightly", "2025-01-06"),
        ("RETAIL1H26", "semimonthly", "2026-03-15"),
        ("RETAIL2H26", "semimonthly", "2026-03-30"),
        ("CPIH26", "monthly", "2026-03-15"),
        ("GDPH26", "quarterly", "2026-03-30"),
        ("GDPF26", "quarterly", "2026-01-15"),
        ("WORLDGDP2027", "yearly", "2027-03-15"),
    ],
    ids=str,
)
def test_validate_symbol__pass(symbol, frequency, release_date):
    validators.validate_symbol(symbol, frequency, date.fromisoformat(release_date))


@pytest.mark.parametrize(
    "symbol,frequency,release_date",
    [
        ("BTCUSDM2024", "monthly", "2024-06-16"),
        ("BTCUSDA24", "monthly", "2024-01-23"),
        ("CLAIMS54W25", "weekly", "2025-12-31"),
        ("BTCUSD32M24", "daily", "2024-06-31"),
        ("BTCUSDM", "yearly", "2024-06-01"),
        ("WORLDGDP27", "yearly", "2027-12-31"),
        ("BTCUSD15M24", "daily", "2024-06-16"),
        ("CLAIMS02W25", "weekly", "2025-01-23"),
        ("CPIH26", "monthly", "2026-05-15"),
        ("GDPH26", "monthly", "2026-04-15"),
    ],
    ids=str,
)
def test_validate_symbol__error(symbol, frequency, release_date):
    with pytest.raises(ValueError):
        validators.validate_symbol(symbol, frequency, date.fromisoformat(release_date))



import re
from decimal import Decimal
from unittest.mock import Mock

import pytest
import eth_account
from hexbytes import HexBytes
from web3 import Web3, HTTPProvider
from web3.contract.contract import ContractFunction
from web3.eth import Eth

import afp
from afp.api.base import BaseAPI, ClearingSystemAPI, ExchangeAPI
from afp.bindings import MarginAccount
from afp.constants import defaults
from afp.dtos import ExchangeParameters
from afp.exceptions import ConfigurationError
from afp.exchange import ExchangeClient

from . import NULL_ADDRESS, AuthenticatorStub


def test_BaseAPI__raises_error_for_missing_authenticator():
    app = afp.AFP()
    with pytest.raises(ConfigurationError):
        BaseAPI(app.config)


def test_ClearingSystemAPI__raises_error_for_missing_rpc_url():
    app = afp.AFP(authenticator=AuthenticatorStub())
    with pytest.raises(ConfigurationError):
        ClearingSystemAPI(app.config)


def test_ClearingSystemAPI__applies_transaction_parameters(monkeypatch):
    mock_build_transaction = Mock(return_value={})
    mock_wait_for_transaction_receipt = Mock(return_value={})
    mock_get_transaction_count = Mock(return_value=5)
    mock_send_raw_transaction = Mock(return_value=HexBytes("0x"))

    monkeypatch.setattr(ContractFunction, "build_transaction", mock_build_transaction)
    monkeypatch.setattr(Eth, "get_transaction_count", mock_get_transaction_count)
    monkeypatch.setattr(Eth, "send_raw_transaction", mock_send_raw_transaction)
    monkeypatch.setattr(
        Eth, "wait_for_transaction_receipt", mock_wait_for_transaction_receipt
    )

    w3 = Web3()
    margin_contract = MarginAccount(w3, NULL_ADDRESS)
    authorize = margin_contract.authorize(NULL_ADDRESS)

    assert defaults.TIMEOUT_SECONDS != 4

    app = afp.AFP(
        authenticator=AuthenticatorStub(),
        rpc_url="http://foobar",
        gas_limit=1,
        max_fee_per_gas=2,
        max_priority_fee_per_gas=3,
        timeout_seconds=4,
    )
    api = ClearingSystemAPI(app.config)

    assert isinstance(api._w3.provider, HTTPProvider)
    assert api._w3.provider.endpoint_uri == "http://foobar"

    api._transact(authorize)

    mock_build_transaction.assert_called_once()
    tx_params = mock_build_transaction.call_args_list[0].args[0]
    assert tx_params["from"] == NULL_ADDRESS
    assert tx_params["nonce"] == 5
    assert tx_params["gasLimit"] == 1
    assert tx_params["maxFeePerGas"] == 2
    assert tx_params["maxPriorityFeePerGas"] == 3

    mock_wait_for_transaction_receipt.assert_called_once()
    timeout = mock_wait_for_transaction_receipt.call_args_list[0].kwargs["timeout"]
    assert timeout == 4


def test_ExchangeAPI__applies_exchange_parameters(monkeypatch):
    fake_exchange_params = ExchangeParameters(
        trading_protocol_id="baz",
        maker_trading_fee_rate=Decimal("0"),
        taker_trading_fee_rate=Decimal("0"),
    )
    mock_login = Mock(return_value=fake_exchange_params)
    mock_generate_login_nonce = Mock(return_value="12345678")

    monkeypatch.setattr(ExchangeClient, "login", mock_login)
    monkeypatch.setattr(
        ExchangeClient, "generate_login_nonce", mock_generate_login_nonce
    )

    app = afp.AFP(
        authenticator=AuthenticatorStub(), exchange_url="http://foobar", chain_id=12345
    )
    api = ExchangeAPI(app.config)

    assert api._exchange._base_url == "http://foobar"
    mock_login.assert_called_once()
    login_submission = mock_login.call_args_list[0].args[0]
    assert "Chain ID: 12345" in login_submission.message


def test_ExchangeAPI__generates_eip4361_message(monkeypatch):
    nonce = "12345678"
    account = eth_account.Account.from_key(
        "0x32df57bd2cbdca044227974f6937d5722da13344218daa4286071e7850d28694"
    )

    monkeypatch.setattr(ExchangeAPI, "_login", Mock())

    expected_message_regex = re.compile(
        r"\S+ wants you to sign in with your Ethereum account:\n"
        rf"{account.address}\n\n\n"
        r"URI: \S+\n"
        r"Version: 1\n"
        r"Chain ID: \S+\n"
        rf"Nonce: {nonce}\n"
        r"Issued At: \S+"
    )

    app = afp.AFP(authenticator=afp.PrivateKeyAuthenticator(account.key))
    actual_message = ExchangeAPI(app.config)._generate_eip4361_message(nonce)

    assert expected_message_regex.match(actual_message)



"""Typed bindings around the smart contracts of the Autonomous Futures Protocol."""

from .facade import (
    ClearingDiamond,
    MarginAccountRegistry,
    OracleProvider,
    ProductRegistry,
    SystemViewer,
)
from .margin_account import MarginAccount
from .types import (
    BaseProduct,
    ClearingConfig,
    Config,
    ExpirySpecification,
    FinalSettlementConfig,
    FuturesProductV1,
    Intent,
    IntentData,
    MarginData,
    MarginSpecification,
    MarkPriceConfig,
    OracleSpecification,
    PositionData,
    PredictionProductV1,
    ProductConfig,
    ProductMetadata,
    ProductState,
    Settlement,
    Side,
    Trade,
)

__all__ = (
    # Contract bindings
    "ClearingDiamond",
    "MarginAccount",
    "MarginAccountRegistry",
    "OracleProvider",
    "ProductRegistry",
    "SystemViewer",
    # Structures
    "BaseProduct",
    "ClearingConfig",
    "Config",
    "ExpirySpecification",
    "FinalSettlementConfig",
    "FuturesProductV1",
    "Intent",
    "IntentData",
    "MarginData",
    "MarginSpecification",
    "MarkPriceConfig",
    "OracleSpecification",
    "PositionData",
    "PredictionProductV1",
    "ProductConfig",
    "ProductMetadata",
    "ProductState",
    "Settlement",
    "Side",
    "Trade",
    # Enumerations
    "ProductState",
    "Side",
)



"""MarkPriceTrackerFacet contract binding and data structures."""

# This module has been generated using pyabigen v0.2.16

import typing

import eth_typing
import hexbytes
import web3
from web3 import types
from web3.contract import contract


class MarkPriceTrackerFacet:
    """MarkPriceTrackerFacet contract binding.

    Parameters
    ----------
    w3 : web3.Web3
    address : eth_typing.ChecksumAddress
        The address of a deployed MarkPriceTrackerFacet contract.
    """

    _contract: contract.Contract

    def __init__(
        self,
        w3: web3.Web3,
        address: eth_typing.ChecksumAddress,
    ):
        self._contract = w3.eth.contract(
            address=address,
            abi=ABI,
        )

    def valuation(
        self,
        product_id: hexbytes.HexBytes,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `valuation` on the MarkPriceTrackerFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.valuation(
            product_id,
        ).call(block_identifier=block_identifier)
        return int(return_value)

    def valuation_after_trade(
        self,
        product_id: hexbytes.HexBytes,
        price: int,
        quantity: int,
        block_identifier: types.BlockIdentifier = "latest",
    ) -> int:
        """Binding for `valuationAfterTrade` on the MarkPriceTrackerFacet contract.

        Parameters
        ----------
        product_id : hexbytes.HexBytes
        price : int
        quantity : int
        block_identifier : web3.types.BlockIdentifier
            The block identifier, defaults to the latest block.

        Returns
        -------
        int
        """
        return_value = self._contract.functions.valuationAfterTrade(
            product_id,
            price,
            quantity,
        ).call(block_identifier=block_identifier)
        return int(return_value)


ABI = typing.cast(
    eth_typing.ABI,
    [
        {
            "type": "function",
            "name": "valuation",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"}
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {
            "type": "function",
            "name": "valuationAfterTrade",
            "inputs": [
                {"name": "productId", "type": "bytes32", "internalType": "bytes32"},
                {"name": "price", "type": "int256", "internalType": "int256"},
                {"name": "quantity", "type": "uint256", "internalType": "uint256"},
            ],
            "outputs": [{"name": "", "type": "int256", "internalType": "int256"}],
            "stateMutability": "view",
        },
        {"type": "error", "name": "EVWMA_NotInitialized", "inputs": []},
        {
            "type": "error",
            "name": "InvalidFieldAccess",
            "inputs": [
                {
                    "name": "productType",
                    "type": "uint8",
                    "internalType": "enum ProductType",
                },
                {"name": "field", "type": "string", "internalType": "string"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv18_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_MulDiv_Overflow",
            "inputs": [
                {"name": "x", "type": "uint256", "internalType": "uint256"},
                {"name": "y", "type": "uint256", "internalType": "uint256"},
                {"name": "denominator", "type": "uint256", "internalType": "uint256"},
            ],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Div_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Div_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp2_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Exp_InputTooBig",
            "inputs": [{"name": "x", "type": "int256", "internalType": "SD59x18"}],
        },
        {"type": "error", "name": "PRBMath_SD59x18_Mul_InputTooSmall", "inputs": []},
        {
            "type": "error",
            "name": "PRBMath_SD59x18_Mul_Overflow",
            "inputs": [
                {"name": "x", "type": "int256", "internalType": "SD59x18"},
                {"name": "y", "type": "int256", "internalType": "SD59x18"},
            ],
        },
    ],
)
